# VSS2SVN.py
# 
# Purpose:
#   To convert a VSS repository to an svn repository. Actually, what it does
# is iterate over a VSS repository and ultimately creates an svn dump file that
# can subsequently be loaded into an svn repository.
#
## ================================================================
## Copyright (c) 2000-2003 CollabNet.  All rights reserved.
## 
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are
## met:
## 
## 1. Redistributions of source code must retain the above copyright
## notice, this list of conditions and the following disclaimer.
## 
## 2. Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3. The end-user documentation included with the redistribution, if
## any, must include the following acknowledgment: "This product includes
## software developed by CollabNet (http://www.Collab.Net/)."
## Alternately, this acknowledgment may appear in the software itself, if
## and wherever such third-party acknowledgments normally appear.
## 
## 4. The hosted project names must not be used to endorse or promote
## products derived from this software without prior written
## permission. For written permission, please contact info@collab.net.
## 
## 5. Products derived from this software may not use the "Tigris" name
## nor may "Tigris" appear in their names without prior written
## permission of CollabNet.
##
## THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
## WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
## MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
## IN NO EVENT SHALL COLLABNET OR ITS CONTRIBUTORS BE LIABLE FOR ANY
## DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
## DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
## GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
## IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##
## ====================================================================
## 
## This software consists of voluntary contributions made by many
## individuals on behalf of CollabNet.
##
## A partial list follows:
## Jens B. Jorgensen    jjorge00    Original author
## Richard C. Balint    archfool    Freelance Hacker
##
import re, os, time, sha, sys, fileinput, tempfile, md5, time, string, types
from win32com.client import Dispatch

#
# Classes and functions
#

# Encapsulates an event that happened in the VSS repository
class RevInfo :
    def __init__(self) :
        # time_t timestamp at which the VSS event occurred. after we build our whole
        # collection of these we will sort by timestamp
        self.timestamp = 0
        # sha-1 hash of author+comment. we use this along with timestamp to coalesce
        # VSS events into svn revisions
        self.hash      = ''
        # This is the VSS revision # on the above VSS node where the event occurred
        self.rev       = 0
        # The VSS Node Path plus '/' for the directory if this event is a directory
        # event of for the directory that owns the file is this is a file event
        self.parent    = ''
        # If self.type == VSSITEM_FILE this is the current basename of the node,
        # if self.type == VSSITEM_PROJECT then it is the basename(s) of the node at the time
        # the event occurred. If this was a rename then it will be a tuple with from/to
        self.filename  = ''
        # This flag indicates that a given file node is a binary file
        self.binary    = 0
        # Either VSSITEM_FILE or VSSITEM_PROJECT according to what kind of node the event
        # occurred on. Note that this is PROJECT for basically everything but file checkin
        self.type      = 1
        self.username  = ''
        self.comment   = ''
        self.action    = ''
        self.deleted   = 0
        self.target    = ''

class Dumper:
  def __init__(self, dumpfile_path):
    'Open DUMPFILE_PATH, and initialize revision to REVISION.'
    self.dumpfile_path = dumpfile_path
    self.revision   = 0
    self.dumpfile   = open(dumpfile_path, 'wb')
    self.dname      = os.path.join(os.environ['TMP'], 'dumptemp') #tempfile.mkdtemp() this does not work in 2.2
    if not os.path.exists(self.dname):
        os.mkdir(self.dname)
    self.latestrevs = {}

    # Initialize the dumpfile with the standard headers:
    #
    # The CVS repository doesn't have a UUID, and the Subversion
    # repository will be created with one anyway.  So when we load
    # the dumpfile, we'll tell svnadmin to ignore the UUID below. 
    self.dumpfile.write('SVN-fs-dump-format-version: 2\n'
                        '\n'
                        'UUID: ????????-????-????-????-????????????\n'
                        '\n')

  def start_revision(self, props):
    """Write the next revision, with properties, to the dumpfile.
    Return the newly started revision."""

    self.revision = self.revision + 1

    if not quiet and not debug and (self.revision % 10) == 0 :
        print "Creating subversion revision: ", self.revision
    
    # A revision typically looks like this:
    #
    #   Revision-number: 1
    #   Prop-content-length: 129
    #   Content-length: 129
    #   
    #   K 7
    #   svn:log
    #   V 27
    #   Log message for revision 1.
    #   K 10
    #   svn:author
    #   V 7
    #   jrandom
    #   K 8
    #   svn:date
    #   V 27
    #   2003-04-22T22:57:58.132837Z
    #   PROPS-END
    #
    # Notice that the length headers count everything -- not just the
    # length of the data but also the lengths of the lengths, including
    # the 'K ' or 'V ' prefixes.
    #
    # The reason there are both Prop-content-length and Content-length
    # is that the former includes just props, while the latter includes
    # everything.  That's the generic header form for any entity in a
    # dumpfile.  But since revisions only have props, the two lengths
    # are always the same for revisions.
    
    # Calculate the total length of the props section.
    total_len = 10  # len('PROPS-END\n')
    for propname in props.keys():
      klen = len(propname)
      klen_len = len('K %d' % klen)
      vlen = len(props[propname])
      vlen_len = len('V %d' % vlen)
      # + 4 for the four newlines within a given property's section
      total_len = total_len + klen + klen_len + vlen + vlen_len + 4
        
    # Print the revision header and props
    self.dumpfile.write('Revision-number: %d\n'
                        'Prop-content-length: %d\n'
                        'Content-length: %d\n'
                        '\n'
                        % (self.revision, total_len, total_len))

    for propname in props.keys():
      self.dumpfile.write('K %d\n' 
                          '%s\n' 
                          'V %d\n'
                          '%s\n' % (len(propname),
                                    propname,
                                    len(props[propname]),
                                    props[propname]))

    self.dumpfile.write('PROPS-END\n\n')

    return self.revision

  def add_dir(self, path):
    if len(path) > 0 and not path == '/':
        self.dumpfile.write("Node-path: %s\n"
                            "Node-kind: dir\n"
                            "Node-action: add\n"
                            "Prop-content-length: 10\n"
                            "Content-length: 10\n"
                            "\n"
                            "PROPS-END\n"
                            "\n"
                            "\n" % path)


  def add_or_change_file(self, rev_item):
	if not self.latestrevs.has_key(rev_item.realfilename):
		rev_item.action = 'added'

	self.latestrevs[rev_item.realfilename] = self.revision
	if rev_item.binary :
	  is_executable = 1
	  # "K 14\n" + "svn:executable\n" + "V 1\n" + "*\n" + "PROPS-END\n"
	  props_len = 36
	else:
	  is_executable = 0
	  # just "PROPS-END\n"
	  props_len = 10

	if rev_item.action == 'added' :
		action = 'add'
	else :
		action = 'change'
	    

	# We want to get the version of this file
	# First check the purged list, then try normal file.
	# If that doesn't work, try to get a deleted file
	fullname = rev_item.realfilename

	TempFile = self.dname + '\\' + rev_item.filename
	print "DName: ", TempFile
	Version = rev_item.version
	print Version.VSSItem.spec

	try :
		if rev_item.type == VSSITEM_PROJECT : Version.VSSItem.Get(self.dname)
		else : Version.VSSItem.Get(TempFile)
		f = open(TempFile, 'rb')

		# Insert the rev contents, calculating length and checksum as we go.
		buf = f.read()
		f.close()
		del_file(TempFile)
	except :
		buf = 'File revision could not be retrived from VSS'
		
		
	checksum = md5.new()
	length = 0
	checksum.update(buf)
	length = len(buf)

	self.dumpfile.write('Node-path: %s\n'
						'Node-kind: file\n'
						'Node-action: %s\n'
						'Prop-content-length: %d\n'
						'Text-content-length: %16d\n'
						'Text-content-md5: %s\n'
						'Content-length: %16d\n'
						'\n'
						% (fullname, action, props_len, length,
						   checksum.hexdigest(), (length + props_len)))

	if is_executable:
	  self.dumpfile.write('K 14\n'
						  'svn:executable\n'
						  'V 1\n'
						  '*\n')

	self.dumpfile.write('PROPS-END\n')
	self.dumpfile.write(buf)
	self.dumpfile.write('\n')
		

  def rename_file(self, revitem) :
    oldrev = self.latestrevs[revitem.realfilename]
    fullname = revitem.parent + revitem.filename[0]
    action = revitem.action.split()
    newname = revitem.parent + revitem.filename[1]
    self.latestrevs[newname] = self.revision

    self.dumpfile.write('Node-path: %s\n'
                        'Node-kind: file\n'
                        'Node-action: add\n'
                        'Node-copyfrom-rev: %d\n'
                        'Node-copyfrom-path: %s\n'
                        '\n' % (newname, oldrev, fullname))
    
    self.dumpfile.write('Node-path: %s\n'
                        'Node-kind: file\n'
                        'Node-action: delete\n'
                        '\n' % fullname)


  def del_file(self, revitem) :
    #self.latestrevs[revitem.realfilename] = -1
    fullname = revitem.parent + revitem.filename
    self.dumpfile.write('Node-path: %s\n'
                        'Node-kind: file\n'
                        'Node-action: delete\n'
                        '\n' % fullname)

  def undel_file(self, revitem) :
    oldrev = self.latestrevs[revitem.realfilename]
    fullname = revitem.realfilename
    self.latestrevs[fullname] = self.revision

    self.dumpfile.write('Node-path: %s\n'
                        'Node-kind: file\n'
                        'Node-action: add\n'
                        'Node-copyfrom-rev: %d\n'
                        'Node-copyfrom-path: %s\n'
                        '\n' % (fullname, oldrev, fullname))

  def del_recurse(self, dirname) :
      for i in os.listdir(dirname) :
          fname = os.path.join(dirname, i)
          if os.path.isdir(fname) :
              self.del_recurse(fname)
              os.rmdir(fname)
          else :
              os.chmod(fname, 0666)
              os.unlink(fname)
      os.rmdir(dirname)
      
  def close(self):
    if debug : print "\nDumper.close(): Removing temporary directory: ", self.dname
    self.del_recurse(self.dname)
    self.dumpfile.close()


def rev_infos_insert(ri) :
    low = 0
    high = len(rev_infos) - 1
    pvt = 0
    while low < high :
        pvt = (low + high) / 2
        if ri.timestamp < rev_infos[pvt] :
            low = pvt + 1
        elif ri.timestamp > rev_infos[pvt] :
            high = pvt - 1
        else :
            break
    rev_infos.insert(pvt, ri)

def del_file(fname) :
    for i in range(0, 10) :
        try :
            os.chmod(fname, 0666)
            os.unlink(fname)
            return
        except :
            if debug : print "Can't remove file %s. Sleeping" % fname
            time.sleep(0.2)
    raise RuntimeError('couldn\'t delete %s' % fname)



def format_date(date):
  """Return an svn-compatible date string for DATE (seconds since epoch)."""
  # A Subversion date looks like "2002-09-29T14:44:59.000000Z"
  return time.strftime("%Y-%m-%dT%H:%M:%S.000000Z", time.gmtime(int(date)))

def get_version(vssitem, revnum) :
    # unfortunately there is not way to get a VSSItem to give us
    # a VSSVersion according to a version number so we have to iterate
    # through them to find the one we want
    for v in vssitem.Versions :
        print "Get_Version: ", vssitem.spec, v.VersionNumber, revnum
        # labels don't have their own version number, they just occur at a point
        # in time but they do have a VersionNumber attached to them
        if v.Action[0:7] == 'Labeled' :
            continue
        if v.VersionNumber == revnum :
            return v
    return None





def printhelp() :
    print
    print "vss2svn.py command-line options: (Options with * do not take arguments)"
    print "  Short    Long      Use                          Default Value"
    print "  -----   ------     ---------------------------  -------------"
    print "   -h     -help      * Print this help            N/A"
    print "   -r     -repos     Path to srcsafe.ini          $SSDIR"
    print "   -s     -ss        Path to ss.exe               C:\Program Files\Microsoft Visual Studio\Common\VSS\win32"
    print "   -u     -user      VSS Username                 $SSUSER"
    print "   -p     -pass      VSS Password                 (blank)"
    print "   -i     -proj      VSS Project                  $/"
    print "   -o     -dump      Subversion dumpfile          $TEMP/svn_dump"
    print "   -t     -thresh    Commit threshold in minutes  5 minutes"
    print "   -d     -debug     * Turn on verbose output     Off"
    print "   -q     -quiet     * Less output than normal    Off"
    print "   -x     -exclude   VSS Sub-projects to exclude  (none)"
    print "   -n     -nohistory VSS project tree to only"
    print "                     the latest file versions     (none)"
    print "   -a     -argfile   argument file                (none)"
    print "                     Use multiple -x for multiple"
    print "                     excludes"

# NOTE: originally this script was written to utilize the VSS Automation
# interface so that all this disgusting parsing of command-line output
# would not have to be done. Unfortunately, the VSS Automation API
# is crippled in some dire and more unfortunately, unfixable ways. You
# can get around some of the issues. Anyhow so the command-line is used
# to get histories on the files. For a nice in-depth read of what is wrong
# with VSS (including but not limited to why we must use the command-line
# tool) read http://www.michaelbolton.net/testing/VSSDefects.html

ver_re = re.compile('.*Version (\d+).*')
user_date_re = re.compile('.*User: (\\S+)\\s+Date:\\s+(\\S+)\\s+Time:\\s+(\\S+).*')
suffix_action_re = re.compile('^(.+) (added|deleted|purged|destroyed|recovered|shared|branched|Rolled back|unpinned)$')
renamed_re = re.compile('^(.+) renamed to (.+)$')
moved_re = re.compile('^(.+) moved to (.+)$')
pinned_re = re.compile('^(.+) pinned to (.+)$')
fileheader_re = re.compile('^  ((\\S| )+)  \\*\\*\\*\\*\\*$')
checkin_re = re.compile('^Checked in (.+)$')
branched_re = re.compile('^Branched$')
rolledback_re = re.compile('^Rolled back$')

def parse_action(l,l2) :
    "returns (action, filename(s))"
    m = suffix_action_re.match(l)
    if m :
        return (m.group(2), m.group(1))
    m = renamed_re.match(l)
    if m :
        return ('renamed', (m.group(1), m.group(2)))
    m = moved_re.match(l)
    if m :
        return ('moved to', (m.group(1), m.group(2)))
    m = pinned_re.match(l)
    if m :
        return ('pinned', (m.group(1), m.group(2)))
    m = checkin_re.match(l)
    if m :
        return ('checked', m.group(1))
    m = branched_re.match(l)
    if m :
        return ('branched', '')
    m = rolledback_re.match(l)
    if m :
        return ('rolled back', '')
    if l == 'Created' :
        return ('created', '')

    if not l2 == None:
        return parse_action(l + ' ' + l2,None)

    raise RuntimeError('unhandled action line: %s' % l)

def to_timestamp(date_str, time_str) :
    "take date like mm/dd/yy and time like HH:MM<a/p> and return time_t"
    (mm, dd, yy) = map(lambda x : int(x), date_str.split('/'))
    if yy < 50 :
        yy += 2000
    else :
        yy += 1900
    m = re.match('(\\d+):(\\d+)(.)', time_str)
    HH = int(m.group(1))
    MM = int(m.group(2))
    a_p = m.group(3)
    if a_p == 'p' :
        if HH != 12 :
            HH += 12
    else :
        if HH == 12 :
            HH = 0
    return time.mktime((yy, mm, dd, HH, MM, 0, 0, 0, 0))

def FormatProject(Project):
    Path = Project

    if Path[0] == '$':
        Path = Path[1:]

    if not Path[-1] == '/':
        Path = Path + '/'

    if not Path[0] == '/':
        Path = '/' + Path

    return Path

def StripNonAscii(Str):
    Result = ''
    for x in range(len(Str)):
        if ord(Str[x]) <= 127:
            Result = Result + Str[x]

    return Result

def GetDirectoryList(startproj) :
    List = [FormatProject(startproj)]

    Node = vssdb.VSSItem(startproj)
    for Item in Node.Items:
        if Item.Type == VSSITEM_PROJECT:
            #print "Project:",FormatProject(startproj) + Item.Name
            List = List + GetDirectoryList(FormatProject(startproj) + Item.Name)

    return List

# To get the full history of all events that happened in a VSS repository
# we run the ss.exe command line tool recursively on each directory
# in the repository. We descend through the tree creating a RevInfo
# object for each event we discover.
def GetInfoCache(startproj) :
    InfoCache = []

    revline = RevInfo()
    revline.parent    = FormatProject(startproj)
    revline.timestamp = 0   # use zero so directories are added first
    revline.type      = VSSITEM_PROJECT
    revline.action    = 'diradded'
    revline.hash      = sha.new("An arbitrary string so all directories are added together").hexdigest()

    InfoCache.append(revline)

    Node = vssdb.VSSItem(startproj)
    for Item in Node.Items:
        if Item.Type == VSSITEM_FILE:
            for ver in Item.Versions :
                if ver.Action[:7] == 'Labeled':
                    continue
                revline = RevInfo()

                revline.comment   = StripNonAscii(ver.Comment)
                revline.parent    = FormatProject(startproj)
                revline.filename  = Item.Name
                revline.username  = ver.UserName
                revline.rev       = int(ver.VersionNumber)
                revline.timestamp = int(ver.Date)
                revline.type      = VSSITEM_FILE
                revline.action    = ver.Action
                revline.hash      = sha.new(revline.comment + '\0' + ver.UserName).hexdigest()
                revline.version   = ver

                InfoCache.append(revline)
        else:
            project = FormatProject(startproj) + Item.Name
            if FormatProject(project.lower()) not in dir_filter:
                print "Project:",project
                InfoCache = InfoCache + GetInfoCache(project)

    return InfoCache

    InfoCache = []
    SubProj   = [startproj]
    Ignoring  = []
    while len(SubProj) :
        ThisProj = SubProj.pop(0)
        user_pass = ''
        if vss_user :
            if vss_passwd :
                user_pass = ' -Y' + vss_user + "," + vss_passwd
            else :
                user_pass = ' -Y' + vss_user
        cmd = '' + ss + ' History ' + '"' + ThisProj + '"' +  user_pass

        print
        print cmd

        SSPipe = os.popen(cmd)
        history = SSPipe.read()
        retval = SSPipe.close()

        if retval != None :
            Ignoring.append(ThisProj.upper())
            print ThisProj + ' not found, skipping'
            continue

        history2 = history.split('\n*****')

        # $/ always gives us problems later on. Chomp off trailing slash
        if ThisProj == '$/' : ThisProj = "$"

        name = ThisProj[2:]
        name = name.replace("/","_")
        T = open("history (" + name + ").out","w")
        #T = open("history.out","w")
        T.write(history)
        T.close()

        # Skip the first line: it contains no useful data
        count = 1
        for line in history2[1:] :
            #print line
            linearr = line.split('\n')
            # the revision info could be talking about a file in this directory
            # or the directory
            m = fileheader_re.match(linearr[0])
            if m :
                if linearr[2][:6] == 'Label:':
                    continue
                filename = m.group(1)
                m = ver_re.match(linearr[1])
                rev = int(m.group(1))
                m = user_date_re.match(linearr[2])
                username = m.group(1)
                rev_date = m.group(2)
                rev_time = m.group(3)

                type = VSSITEM_FILE
                (action, action_fname) = parse_action(linearr[3],linearr[4])
                parent = ThisProj #+ filename
            else :
                print '\r',linearr[0],
                # Skip labels
                if linearr[1][:6] == 'Label:':
                    continue
                m = ver_re.match(linearr[0])
                rev = int(m.group(1))
                m = user_date_re.match(linearr[1])
                username = m.group(1)
                rev_date = m.group(2)
                rev_time = m.group(3)

                type = VSSITEM_PROJECT
                (action, filename) = parse_action(linearr[2],linearr[3])
                linearr = []
                parent  = ThisProj
                # project history blocks can actually describe files
                if len(filename) > 0 and filename[0] != '$':
                    type = VSSITEM_FILE

            # Get any comments
            comment = ''
            for i in range(3,len(linearr)):
                if re.match('^Comment: .*', linearr[i]):
                    c_lines = ''.join(linearr[i:])
                    comment = c_lines[9:]

            count = count + 1

            # Now, we should have three or four general items left.
            #    Filename, Parent, Action [, Target]
            # We just need to figure out which is which. Of course, it's
            # different for each action and item type. =sigh=

            if action[:6] == 'pinned' or action[:8] == 'unpinned' or action[:8] == 'branched' or action[:11] == 'rolled back' or action[:8] == 'shared' or action[:8] == 'moved to' or action[:9] == 'destroyed' or action[:7] == 'deleted' or action[:6] == 'purged':
                continue

            if type == VSSITEM_FILE :
                if action == 'renamed':
                    path = parent + '/' + filename[0]
                elif action == 'added':
                    # added actions can come from directory blocks and then the
                    # version number needs to be reset
                    path = parent + '/' + filename[0]
                    rev = 1
                else:
                    path = parent + '/' + filename
            else :
                if parent == '$' : path = '$/'
                else : path = parent
                # if the filename starts with a $, it's a new path we created
                if len(filename) > 0 and filename[0] == '$' :
                    if action == 'added' :
                        filename = filename[1:]
                        SubProj.append(parent + '/' + filename)
                        continue
                    # Handle directory moving.  We get an entry for both the
                    # from and to directory. Skip one of them.
                    elif action[:10] == 'moved from' :
                        filename = filename[1:]
                        SubProj.append(parent + '/' + filename)
                    else :
                        raise RuntimeError('action %s not handled' % action)

            try:
                Node = vssdb.VSSItem(path)
            except:
                if action == 'added':
                    # VSS failed to get this item, maybe it has been purged?
                    # We will ignore it
                    Ignoring.append(path.lower())
                    continue
                timestamp = to_timestamp(rev_date,rev_time)

            Version = None
            for ver in Node.Versions :
                if ver.VersionNumber == rev and ver.VSSItem.Spec == path :
                    #print 'found %s ver %s' % (path, rev)
                    timestamp = int(ver.Date)
                    Version = ver
                    break

            if not Version :
                Ignoring.append(path.upper() + "(" + str(rev) + ")")
                continue
            #print rev_date, rev_time
            #print time.ctime(timestamp)

            parent = parent + '/'

            # Print out a warning if we cannot find the revision
            if timestamp == -1 :
                print "\tNO TIMESTAMP for: ", path, rev

            # we don't include creation of the root dir since SVN is not
            # going to be happy with that
            if parent == '$/' and action == 'created' and filename == '' :
                continue
            #print parent, action, filename
            digest = sha.new(comment + '\0' + username).hexdigest()

            revline = RevInfo()
            revline.parent    = parent[1:]  # chop off leading '$'
            revline.filename  = filename
            revline.username  = username
            revline.rev       = int(rev)
            revline.timestamp = int(timestamp)
            revline.type      = type
            revline.action    = action
            revline.comment   = comment
            revline.hash      = digest
            revline.version   = Version

            InfoCache.append(revline)

    Ignoring.sort()
    T = open("ignoring.out","w")
    for l in Ignoring:
        T.write(l)
        T.write('\n')
    T.close()

    return InfoCache


#
# Begin processing / Main Program
# 

#
# Constants
#

# Boolean 'Constants'
FALSE = 0
TRUE  = 1

VSS_Actions = ('added',   'checked', 'branched', 'moved',   'recovered',
               'renamed', 'labeled', 'shared',   'deleted', 'purged',
               'created', 'destroyed')

# VSS Constants
# (ref http://msdn.microsoft.com/library/en-us/dnvss/html/vssauto.asp)
VSSITEM_PROJECT = 0
VSSITEM_FILE    = 1
VSSFILE_CHECKEDOUT_ME = 2

# VSS OLE CLSID
# (ref http://support.microsoft.com/default.aspx?scid=kb;en-us;q201431)
VSS_Dispatch = '{783CD4E4-9D54-11CF-B8EE-00608CC9A71F}'


#
# Variable Defaults
#
debug       = 1
quiet       = 0
dump_file   = os.environ['TEMP'] + '\\svn_dump'
vss_repos   = None
if os.environ.has_key('SSDIR') :
    vss_repos = os.environ['SSDIR']
vss_user    = None
if os.environ.has_key('SSUSER') :
    vss_user = os.environ['SSUSER']
vss_passwd  = None
vss_base    = '$/'
dir_filter  = []
NoHistory_filter = []
sspath      = 'C:\\Program Files\\Microsoft Visual Studio\\Common\\VSS\\win32\\'
#sspath      = 'c:\\vss\\win32\\'
# Commit threshold is default to 5 minutes, except in debug mode, in which case,
# it's 2 minutes
if not debug : COMMIT_THRESHOLD = 5 * 60
else : COMMIT_THRESHOLD = 2 * 60

def poparg():
  arg = argv.pop(0)
  arg = arg.replace("%20"," ")
  return arg

#
# Parse command line options
#
argv    = sys.argv
vss2svn = poparg()
while len(argv) :
    arg = poparg()

    if arg[:2] == '-h' :
        printhelp()
        sys.exit(0)
    elif arg[:2] == '-d' :
        debug = 1
    elif arg[:2] == '-q' :
        quiet = 1
        debug = 0
    elif arg[:2] == '-r' :
        vss_repos = poparg() # + '\\srcsafe.ini'
    elif arg[:2] == '-u' :
        vss_user= poparg()
    elif arg[:2] == '-p' :
        vss_passwd = poparg()
    elif arg == '-i' or arg[:5] == '-proj' :
        vss_base = poparg()
    elif arg == '-o' or arg[:5] == '-dump' :
        dump_file = poparg()
    elif arg[:2] == '-s' :
        sspath = poparg()
    elif arg[:2] == '-t' :
        COMMIT_THRESHOLD = int(poparg()) * 60
    elif arg == '-x' or arg == '-exclude' :
        dir_filter.append(FormatProject(poparg()).lower())
    elif arg == '-n' or arg == '-nohistory' :
        NoHistory_filter.append(FormatProject(poparg()).lower())
    elif arg == '-a' or arg == '-argfile' :
        try:
            f = open(poparg(),"r")
            data = f.read()
            f.close()
            argv = argv + data.split()
        except:
            pass
    else :
        print "UNKNOWN ARGUMENT: ", arg
        printhelp()
        sys.exit(1)

ss = os.path.join(sspath, "ss.exe")

# We do some sanity checking to make sure the user has supplied all that we need (as near as we
# can tell)

if not vss_repos :
    print 'You must specify the repository, either by setting the environment variable SSDIR or by passing it on the command-line with -r'
    sys.exit(1)

vss_repos_ini = os.path.join(vss_repos, 'srcsafe.ini')
if not os.path.exists(vss_repos_ini) :
    print 'The file %s does not exist. Your VSS repository directory should contain the file srcsafe.ini' % vss_repos_ini
    sys.exit(1)

try :
    os.stat(ss)
except :
    print 'The path to the VSS command-line program is not valid (%s), you\'ll need to modify ss_path in this script' % ss
    sys.exit(1)

    
if not quiet :
    print
    print "VSS Repository:     ", vss_repos
    print "VSS User:           ", vss_user
    if vss_passwd :
        print "VSS Password:       *"+vss_passwd+"*"
    print "VSS Base:           ", vss_base
    print "Dump File:          ", dump_file
    print "Commit Threshold:   ", COMMIT_THRESHOLD / 60, "minutes"
    print "Path to SS.EXE:     ", sspath
    print "Excluded projects:  ", dir_filter
    print "No history projects:", NoHistory_filter
    if debug : print "Debug Mode is ON"
    print

os.environ['SSDIR'] = vss_repos

#
# MAIN PROGRAM
#
t1 = time.time()
if not quiet : print "Pre-Parsing VSS Database: %s.\n" % vss_repos

vssdb = Dispatch(VSS_Dispatch)
if vss_passwd :
    vssdb.Open(vss_repos_ini, vss_user, vss_passwd)
else :
    vssdb.Open(vss_repos_ini)
    
# Create transaction list by iterating through the repository
t2 = time.time()

#if debug :
#    Directories = GetDirectoryList("$/")
#    rfile = open("C:\\Temp\\Directories.txt", 'w')
#    for Dir in Directories :
#        rfile.write("%s\n" % (Dir))
#    rfile.close()


InfoCache = GetInfoCache(vss_base)

t3 = time.time()


def mycmp(a, b) :
    m = cmp(a.timestamp, b.timestamp)
    if m == 0 :
        n = cmp(a.parent, b.parent)
        if n == 0 :
            # if time and parents are the same, sort by directory or file
            # operations next
            o = cmp(a.type, b.type)
            if o == 0:
                # if the file operations are the same then go to filename.
                p = cmp(a.filename, b.filename)
                if p == 0:
                    # unbelievable as it is, if everything else is the same,
                    # there's GOT to be a difference in the rev number!
                    return cmp(a.rev, b.rev)
                else : return p
            else : return o
        else : return n
    else : return m

InfoCache.sort(mycmp)

if debug :
    rfile = open("C:\\Temp\\revitems.txt", 'w')
    for rev_item in InfoCache :
            rfile.write("%d %3d %-25s %-15s %d %-8s %-8s '%s'\n" %
                        (rev_item.timestamp, rev_item.rev,
                        rev_item.parent, rev_item.filename, rev_item.type,
                        rev_item.username, rev_item.action, rev_item.comment))
    rfile.close()
pass

#sys.exit(0)

last_t = 0
last_hash = ''
dumper = Dumper(dump_file)

t4 = time.time()
if not quiet :
    print "-----------------------------"
    print "Creating Subversion dump file"
    print "-----------------------------\n"
pass

    
revision = 0
for rev_item in InfoCache :
    # Find the real filename by going through all renames
    if isinstance(rev_item.filename, (types.ListType, types.TupleType)) :
        rev_item.realfilename = rev_item.parent+rev_item.filename[0]
    else :
        rev_item.realfilename = rev_item.parent+rev_item.filename

    # check if it is time to create a new revision
    t = rev_item.timestamp
    if rev_item.hash != last_hash or (t - last_t) > COMMIT_THRESHOLD :
        # ok, time to create a revision
        props = {}
        props['svn:date'] = format_date(rev_item.timestamp)
        props['svn:author'] = str(rev_item.username)
        cmmnt = str(rev_item.comment)
        if len(cmmnt) > 0 :
            props['svn:log'] = cmmnt
            
        if debug : print '\nStarting revision %d with %s' % (revision+1, props)
        revision = dumper.start_revision(props)
        
        last_t = t
        last_hash = rev_item.hash

    #print "\nvss filename", rev_item.filename, "   real filename", rev_item.realfilename, "   time", rev_item.timestamp
    # is this a directory?
    action_arr = rev_item.action.lower().split()
    if rev_item.type == VSSITEM_PROJECT :
        if action_arr[0] == 'diradded' or action_arr[0] == 'created' :
            if debug : print '\tCreated %s' % rev_item.realfilename
            dumper.add_dir(rev_item.parent)
        elif action_arr[0] == 'added' :
            if debug : print '\tAdded %s' % rev_item.realfilename
            dumper.add_or_change_file(rev_item)
        elif action_arr[0] == 'destroyed' :
            if debug : print '\tDestroyed file %s (nothing to do)' % rev_item.realfilename
        elif action_arr[0] == 'deleted' :
            if debug : print '\tDeleted file %s' % rev_item.realfilename
            dumper.del_file(rev_item)
        elif action_arr[0] == 'shared' :
            if debug : print '\tFile sharing not handled by vss2svn.py (yet)'
            continue
        elif action_arr[0] == 'recovered' :
            if debug : print '\tRecovering file: %s' % rev_item.realfilename
            dumper.undel_file(rev_item)
        elif action_arr[0] == 'renamed' :
            if debug :
                newname = rev_item.action.split()
                print '\tRenamed: %s --> %s' % (rev_item.realfilename, newname[-1])
            dumper.rename_file(rev_item)
        elif action_arr[0] == 'purged' :
            continue
        else :
            print "Failed action : ",action_arr[0]
            raise RuntimeError('*** Project: Cannot handle (%s)(%s) ***' %
                               (rev_item.parent, rev_item.action))
    else :
        dumper.add_or_change_file(rev_item)

dumper.close()
if not quiet :
    print "\nDone creating subversion dump file."
    print "\tTotal number of revisions:\t", revision
    print

t5 = time.time()

if not quiet :
    print "Timings:"
    print "\tInitialization: ", t2-t1
    print "\tIterate Repos:  ", t3-t2
    print "\tSort Time:      ", t4-t3
    print "\tCreate Dump:    ", t5-t4
    print "\tTOTAL Time:     ", t5-t1
    print
    print "*** Finished dumping VSS Database ***"
    print



sys.exit(0)



